"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                    FUNCTIONAL PROGRAMMING                                   ||
      ||                        LIST COMPREHENSIONS, MAPS, FILTERS, and LAMBDAS                      ||
      ||                     (faster, shorter, and lazier, but more complex code)                    ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   A list comprehension is a compact method of creating a new list from an old list by applying a function-and
-if-and-maybe-an-else to each element in the old list. Do not use a for loop when a list comprehension can do
the same task. The comprehension does NOT modify the original list, unless the result is assigned back to the
original list identifier. The speed of a list comprehension is roughly twice that of a for loop. The list
comprehension also offers the functionality of both the map instruction and the filter instruction.

   The general form of a list comprehension is that of a backwards for loop. Note that there are no commas
with a list comprehension. Remember, a list must already exist to use a list comprehension. Here are some
examples.

        dataList = [x for x in 'ABCD']                         # data = ['A', 'B', 'C', 'D']
        sqrList  = [x*2 for x in range(1,10) if x % 2 == 0]    # newlist = [4, 8, 12, 16]
        shortLst = [0 if 3 < x < 6 else x for x in range(1,7)] # shortLst: [1, 2, 3, 0, 0, 6]

   A lazy construct (a generator in this section) produces a value only when requested. If you replace the
outer brackets of a list comprehension with parentheses, then the object becomes a generator, not a list.
#-------------------------------------------------------------------------------------------------------------

        A = [1,2]
        B = [5,6,7]
#---Nested for loops:
        L = []
        for x in A:
            for y in B:
                L.append((x,y))  #  L = [(1, 5), (1, 6), (1, 7), (2, 5), (2, 6), (2, 7)]

#---Nested listed comprehension (translation of above):
        L = [(x,y) for x in A
                   for y in B]   #  L = [(1, 5), (1, 6), (1, 7), (2, 5), (2, 6), (2, 7)]

#---Building a matrix with all zeroes.
    A = [ [0 for col in range(3)]
               for row in range(4) ]                      # A = [[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0]]

#---Building a matrix with all zeroes and 1s on a main diagonal.
    A = [ [1 if col == row else 0 for col in range(3) ]
                                  for row in range(4) ]   # A = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [0, 0, 0]]

#---Building a matrix like this: [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10, 11]].
    A = [ [col + 3*row for col in range(3) ]
                                  for row in range(4) ]   # A = [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10, 11]]
#---Extracting a column from a matrix.
    A = [[0,1,2,3],
         [4,5,6,7],
         [8,9,7,6],]
    B = [row[1] for row in A]
    print(B)                             # output: [1, 5, 9]

#---WARNING: Lists create their own internal variables.
    x = 1
    for x in range(3):
        print(x)
    print ('x =', x)          # x = 2 (expected)
    x = 1
    y = [x for x in range(3)] # This is a new x.
    print ('x =', x)          # x = 1 (Yikes!, and NOT expected)
#-------------------------------------------------------------------------------------------------------------

        dataList = [x for x in 'ABC'] # This is a list comprehension. Note the brackets.
        print(dataList)               # ['A', 'B', 'C', 'D']

        dataList = (x for x in 'ABC') # This is a generator. Note the parentheses.
        print(dataList)               # <generator object <genexpr> at 0x023BB788>

        print(next(dataList))         # A [This will also work: print(dList().__next__()).]
        print(next(dataList))         # B
        print(next(dataList))         # C
#       print(next(dataList))         # Error: StopIteration [, because the generator is used up.]

        print(dList())                # <generator object dList at 0x0220EA30>
        print(next(dList()))          # A
        print(next(dList()))          # A <--Surprise, the dList() function is reset.
        print(list(dList())[1])       # B <--This defeats the purpose of using a generator, instead of a list.

        for x in dList():
            print (x, end = ',')      # A,B,C,
        print()
        for x in dList():
            print (x, end = ',')      # A,B,C, The generator is NOT used up.
#-------------------------------------------------------------------------------------------------------------

   Lambda enables you to create a ONE-LINE function on the fly and embed it in-line where you need it. Lambda
functions do not take names or need the word return.  Imagine a one-line function to calculate distance. Of
course, the built-in function hypot() does the same thing.

                        def dist(a,b):
                            return pow(a*a+b*b, 0.5)
                        . . .
                        def main():
                        print(dist(3,4)) # output: 5.0

   Now we will accomplish the same effect with a lambda (= anonymous) function:

    f = (lambda a, b : pow(a*a+b*b, 0.5)) # calculation
    print(f(3,4))                         # output: 5.0

   Here are a few more examples:

    f = (lambda a, b : print(a+b))      # printing
    f(3,4)                              # output: 7

    f = (lambda a, b = 4 : print(a+b))  # default values are allowed
    f(3)                                # output: 7

    f = (lambda a, b : a if a < b else b) # if (ternary type)
    print(f(3, 4))                        # output: 3

    f = (lambda L: list(map(print, L))) # iterating over lists with a map
    f([3,4])                            # output: 3, 4 (listed vertically)

    f = (lambda L: [x*x for x in L])    # iterating over lists with a list comp.
    print(f([3,4]))                     # output: [9, 16]

   Using a lambda function is actually a bit easier than creating-and-naming a standard function. However, the
lambda body is restricted to a one-line instruction, with no (second) colon allowed. In other words, the
lambda functions cannot use for, and while. Even the if must be of the ternary type (without a colon).

   The syntax consists of the word lambda, followed by a few arguments (AKA parameters, variables), followed
by a colon, followed by one highly-restricted line of code, all inside a set of parentheses:

                           (lambda a, b : a+b).

   Why use lambda? The answer is that we rarely need to use a lambda function, and some Python experts feel
that lambdas should never be used. The advantage of using lambda is that the function is physically near the
code that uses the function. Lambdas are used to a small advantage when making a GUI button. Nested lambdas
are possible, but should never be used.
#-------------------------------------------------------------------------------------------------------------


    A filter returns an iterator (a potential sequence that must be cast to a list) consisting of those items
from a given sequence for which function(item) is true. Both filters and maps do not modify the sequences they
are given.

    A map calls function(item) for each of a given sequence's items and returns an iterator (a potential
sequence that must be cast to a list) of the return values. But note well:

                           list[map(function s)]

is equivalent to the list comprehension

                        [function(elt) for elt in s].

    A mapping is a weak form of a list comprehension, and slightly slower. If the function is a one-liner,
then lambda (see below) may be used with map or with a list comprehension.

        L = [1,3,5,7,]
        M = list(map((lambda x:x+3), L)) # output: [4, 6, 8, 10]
        M = [x+3 for x in L]             # output: [4, 6, 8, 10], easier.

        L = [1,3,5,7,]
        M = [0,1,2,3,]
        N = list(map(pow, L, M))                 # output: [1, 3, 25, 343]
        N = list(map((lambda x, y: x**y), L, M)) # output: [1, 3, 25, 343]
        N = [ pow(L[n], M[n]) for n in range(4)] # output: [1, 3, 25, 343]

    Both filter and map apply a function to every element of a list.
       1) The filter returns only the true     elements, without affecting the
          original list.
       2) The map    returns all  the modified elements, without affecting the
          original list.

#---Example 1 (filter)
    def f(x):
        return x % 2 != 0 and x % 3 != 0

    B = list(range(2, 25))
    A = list(filter(f, B))
    print (A)                          # output: [5, 7, 11, 13, 17, 19, 23]

#---Example 2 (map)
    def f(x):
        return x*2

    A = list(map(f, range(1, 5)))
    print (A)                          # output: [2, 4, 6, 8]

#---Example 3 (map)
    A = range(4)
    def f(x, y):
        return x+y

    A = list(map(f, A, A))
    print (A)                          # output: [0, 2, 4, 6, ]

#---Example 4 (map)
    print (list (map(ord, 'ABC')))     # output: [65, 66, 67]

#---Example 5 (map)
    print (list (map(ord, 'ABC')))     # output: [65, 66, 67]
