"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                            LOOPS                                            ||
      ||                                       (for and while)                                       ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   There are only two kinds of loops in Python: the for loop, and the while loop. The important instructions
break and continue both work for both loops.

   Because both the for loop and the while loop headers all require terminating colons, you cannot
accidentally place a semicolon at the end of a loop header and induce an error. This is a tough bug to find in
other languages, but cannot occur in Python.

                        for n in range (3):         +-----------------------+
                            print (n)               |  The output for both  |
                                                    |  loops is the same:   |
                        n = 0                       |          0            |
                        while n < 3:                |          1            |
                            print (n)               |          2            |
                            n += 1                  +-----------------------+

   IMPORTANT: Above, the value of n AFTER the for loop is 2, but the value of n AFTER the while loop is 3.
Note that "for i in range(n, n): print(i)" gives no output.

   The range function has a hidden default step of 1. But we can make it visible and change it:

                    for n in range (7, 1, -2):
                        print(n)  # output: 7, 5, 3

   You can NOT corrupt the for-loop index:

                    for n in range (1, 5):
                        n+=100
                    print (n, end =", ") # output: 101, 102, 103, 104,

   The implication of this is that you can reuse the same for loop index in nested loops (bad idea).

                            for n in range (2, 4):
                                print (n, end =", ")
                                n += 7777
                                for n in range (100, 102):
                                    print ('--', n)
                                    n += 99999

                            # output: 2, -- 100
                            #         -- 101
                            #         3, -- 100
                            #         -- 101

   For loops work nicely with tuples:

                        for a, b in [ (1,2), (3,4), (5,6) ]:
                            print (a) # output: 1,3,5

   WARNING: Beware of changing a list in a for loop body. Why? Because it changes the indexing values for all
of the remaining elements.

                Lst = [8, 7, 6, 5, 1, 2, 4, 7, 9]
                for x in Lst: # <-- ERROR: This should be "for x in Lst[:]:".
                    if  x > 2:
                        Lst.remove(x)
                print(Lst) # output: [7, 5, 1, 2, 7]

   Above, when the 8 (Lst[0]) was removed the loop went to the next element Lst[1], which now is 6, not 7.




   The for and while loops have an optional else statement. If you ever need a flag in a loop, consider using
the loop else instead. A loop's else statement is ALWAYS executed, unless a break is encountered. The
following code fragments do the same thing. Notice the handy function enumerate is also demonstrated.

        #---Method 1. Searching using a flag
            target = 'D'
            Lst = ['A', 'B', 'C', 'D', 'E', 'F']
            flag = True
            for (index, ch) in enumerate(Lst):
                if target == ch:
                    print('Found in position', index)
                    flag = False
                    break
            if flag == True: print('Not found')

        #---Method 2. Searching without a flag.
            target = 'D'
            Lst = ['A', 'B', 'C', 'D', 'E', 'F']
            for (index, ch) in enumerate(Lst):
                if target == ch:
                    print('Found in position', index)
                    break
            else:
                 print('Not found')

        #---Method 3. Using the built-in index function, which raises an error, if the target is not found.
            target = 'G'
            Lst = ['A', 'B', 'C', 'D', 'E', 'F']
            try:
                print('Found in position', Lst.index(target))
            except:
                print('Not found')

   If you need some quick information about a built-in function, just ask for help--e.g., help(list.index),
help(enumerate).
#-------------------------------------------------------------------------------------------------------------

   The range function generates a list of integers. The third parameter (the stride) defaults to 1.

                         range (2, 7)     = [2,3,4,5,6]
                         range (2, 7, 2)  = [2,4,6]
                         range (7, 2, -2) = [7,5,3]

   Surprise, the range function ONLY work with integers. But the for-loop step is most useful for floating
point numbers. So we design a floating-point range function (called frange).

def frange(start, stop, step): # frange = float range.
    if stop <= start: yield None
    i = start
    while i < stop:
        yield i
        i += step

def main():
    from math import pi
    for t in frange (0, 2*pi, 0.5):
        print (t, end = ', ') # output: 0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0,
#-------------------------------------------------------------------------------------------------------------

#   FOR LOOP EXAMPLES.
#---Example 1. A simple loop that prints a final value
    for n in range (4):           # Remember: range (4) = range(0,4) = range (0,4,1) = 0, 1, 2, 3.
        print (n, end = ", ")     # output = 0, 1, 2, 3
    print()
    print (n)                     # output = 3



#---Example 2. using a step value.
    for n in range (1, 6, 2):
        print (n, end = ", ")     # output = 1, 3, 5

#---Example 3. using a step value and going BACKWARDS the wrong way.
    for n in range (1, 6, -2):
        print (n, end = ", ")     # output = (no output)

#---Example 4. using a step value and going BACKWARDS the right way.
    for n in range (6, 1, -2):
        print (n, end = ", ")     # output = 6, 4, 2,

#---Example 5. Sometimes slices can help in loops:
    for x in range(10)[::2]:
        print(x, end = ", ") # output: 0, 2, 4, 6, 8,

#---Example 6. equal values
    for n in range (2,2):
        print (n)                 # output = (no output)

#---Example 7. using continue
    for n in range (4):
        if n % 2 == 0:
           continue
        print (n, end = ", ")     # output = 0, 1, 3

#---Example 8. using break
    for n in range (4):
        if n % 2 == 0:
           continue
        print (n, end = ", ")     # output = 0, 1, 3

#---Example 9. modifying the loop index. (It can't be done.)
    for n in range (0,5):
        n += 100
        print (n, end = ', ')     # output = 100, 101, 102, 103, 104,

#---Example 10. Modifying the range. (This modification CAN be done, and is usually a mistake.)
    Lst = [1, 2, 3, 4, 5, 6,]
    for x in Lst: # This should be "for x in Lst[:]:"--i.e., make a copy of Lst.
          if  x > 2:
            Lst.remove(x)
    print(Lst)                    # output: [1, 2, 4, 6], an unexpected result.

#---Example 11. Modifying the range.
    Lst = [1, 2, 3, 4, 5, 6,]
    for x in Lst[:]:
          if  x > 2:
            Lst.remove(x)
    print(Lst)                    # output: [1, 2], the expected result

#---Example 12. nested loops
    for n in range (1,3):
       for ch in ('A', 'B', 'C'):
           print (ch, end = ',')  # output = A,B,C,; A,B,C,;
       print (';', end = ' ')

#---Example 13. nested loops (Surprise: accidental reuse of index n has no effect.
    for n in range (1,3):
       for n in ('A', 'B', 'C'):
           print (n, end = ',')   # output = A,B,C,; A,B,C,;
       print (';', end = ' ')






#     +-----------------------------------------------------------------------------------------------+
#     | Both Python loops have an ELSE statement, which is rarely used, and poorly named. The else is |
#     | always executed after the loop ends, unless a break has been encountered. Even if the loop    |
#     | body is not executed, the else statement will still be executed. The else is sometimes used   |
#     | to replace flags in for loops.                                                                |
#     +-----------------------------------------------------------------------------------------------+

#---Example 14. else
    for n in range (4):
      print (n, end = ", ")       # output = 0, 1, 2, 3,
      if n == 3: break
    else: (print('ELSE'))

#---Example 15.
    for n in range (4):
      print (n, end = ", ")       # output = 0, 1, 2, 3, ELSE
    else: (print('ELSE'))

#---Example 16. tuples
    T = [(1,2), (3,4), (5,6)]
    for a, b in T: # or "for (a,b) in T:", with parentheses
        print((a,b), end = '')    # output: (1, 2)(3, 4)(5, 6)

#---Example 17. tuples and enumerate (a useful and built-in function, which does not need to be imported.)
    s = ['a', 'b', 'c']
    for n, x in enumerate (s):
        print((n, x), end = ', ') # output = (0, 'a'), (1, 'b'), (2, 'c')
#-------------------------------------------------------------------------------------------------------------

   The for loop is faster than the while loop, but if a while loop is more convenient to code, then use a
while loop. Python does NOT have a goto statement.

#---WHILE LOOP EXAMPLES
#---Example 1.
    a = 2; b = 4
    while a < b:
        print (a, end = ", ")       # output = 2, 3
        a += 1

#---Example 2. break and continue
    a = 2; b = 4
    while a < 4:
      print (a, end = ", ")       # output = 2, 3
      a += 1
      if a == 100: break
      if a ==   2: continue
    print()

#---Example 3. # A kind of "do until" loop                       n = 3
    while (True):
        print(n, end = ' ')   # output: 3 2 1
        n -= 1
        if n == 0: break

#---Example 4. input with a try-except-else statement.
    flag = True
    s = 'Enter an integer:'
    while flag:
     try:
       num = int(input(s))
     except:
       s = 'That was NOT an integer! Try again.\nEnter an integer: '
     else:
       flag = False
#-------------------------------------------------------------------------------------------------------------