"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                               TIMING THE COMPUTER'S EXECUTIONS                              ||
      ||                             timeit's Timer() and time's clock()                             ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   Here is how to time a process:

   from time import clock
   from math import floor
   start = clock()
   sleep(62)
   stop = clock()   # in seconds
   seconds = (stop-start)
   print ("seconds =", seconds)
   print ("Time elapsed =", floor (seconds/60), "minute(s), and", round(seconds%60, 1), "second(s)")
   # output from above: Time elapsed = 1 minute(s), and 2.0 second(s)

   Suppose you needed to randomly generate the two numbers 1 and -1. You find six clever expressions to
randomly generate these numbers. (Are there other ways?). Which is the fastest? We could use the built-in
clock function and obtain the following results:

   1.  r1 = 1; if random() > .5: r1 = -1 # 0.21 milli-seconds
   2.  r2 = int(random() > 0.5) * 2 - 1  # 0.62 milli-seconds
   3.  r3 = int(random()*2)*2 -1         # 0.81 milli-seconds
   4.  r4 = (-1)** int(random()*2)       # 0.98 milli-seconds
   5.  r5 = choice ([1, -1])             # 1.57 milli-seconds
   6.  r6 = randrange(-1, 2, 2)          # 3.26 milli-seconds

   BTW, How do we meaningfully compare the results? Do we say that one expression is a certain percentage
faster or slower than another? No, because percent expressions are open to different interpretations. Instead,
we use ratios:

                The ratio of times of r2/r1 is  2.95.
                The ratio of times of r3/r2 is  1.31.
                The ratio of times of r4/r3 is  1.21.
                The ratio of times of r5/r4 is  1.60.
                The ratio of times of r6/r5 is  2.08.
                The ratio of times of r6/r1 is 15.52.

   Curiously, if this a one-time run, not in a loop, the speed makes no practical difference.

   Python has a built-in timer (called timeit) which contains a Timer() function to time the execution speeds
of particular expressions. The timer runs each expression a million times, and returns the average time for
each execution. You would think the execution time would be constant for each expression. But somehow, it
isn't.

   Which instruction is faster sqrt, pow, or **? We can find out by using the cool timeit module. I am
surprised that sqrt is not the fastest of the three. BTW, different runs give different times. But the ratios
of the times only slightly vary.

         import timeit

         t = timeit.Timer( '100**0.5')
         print (" 1. **    =", round(t.timeit(), 3)) # 1. **    = 0.018 (FAST)

         t = timeit.Timer( 'sqrt(100)',  setup = "from math import sqrt")
         print (" 2. sqrt  =", round(t.timeit(), 3)) # 2. sqrt  = 0.283 (SLOW)

         t = timeit.Timer( 'pow(100, 0.5)',  setup = "from math import pow")
         print (" 3. pow   =", round(t.timeit(), 3)) # 3. pow   = 0.363 (SLOWER)

Below, is a working program with different kinds of examples.
"""
#================================== IMPORTS ===================================
import timeit
from time import clock
#------------------------------------------------------------------------------

#--Warming up (+, -, *, /, **) may be necessary for accurate timing.
t = 137.1+13.2 + 137.1/13.2 + 137.1*13.2 + 137.1**(-13.2)

#--Example 1. Timing add, sub, mult, div, exp, and sqrt.
t = timeit.Timer( '137.1+13.2')
print (" 1. add  =", t.timeit())

t = timeit.Timer( '137.1-13.2')
print (" 2. sub  =", t.timeit())

t = timeit.Timer( '137.1*13.2')
print (" 3. mul  =", t.timeit())

t = timeit.Timer( '137.1/13.2')
print (" 4. div  =", t.timeit())

t = timeit.Timer( '137.1**13.2')
print (" 5. pow  =", t.timeit())

t = timeit.Timer( 'sqrt(13.2)', setup="from math import sqrt")
print (" 6. sqrt =", t.timeit())

t = timeit.Timer( 'random()',   setup="from random import random")
print (" 7. rand =", t.timeit())
#------------------------------------------------------------------------------

#--Example 2. Importing a built-in function. The examples below all randomly generate the two numbers 1 or -1.

t = timeit.Timer( 'int(random()*2)*2 -1',  setup="from random import random")
print (" 8. r1   =", t.timeit()) # Fastest

t = timeit.Timer( '(-1)** int(random()*2)',setup="from random import random")
print (" 9. r2   =", t.timeit()) # 2nd fastest

t = timeit.Timer( 'choice ([1, -1])',      setup="from random import choice")
print ("10. r3   =", t.timeit()) # 3rd fastest

t = timeit.Timer( 'randrange(-1, 2, 2)',   setup="from random import randrange")
print ("11. r4   =", t.timeit()) # 4th fastest
#------------------------------------------------------------------------------

#--Example 3. Working with variables ("pass" means "do nothing" in Python)

x = 1; y = 2; a = 3; b = 4
t = timeit.Timer( "if x==y and a>b: pass", "from __main__ import x, y, a, b")
print ("12. Ans  =", t.timeit())
#------------------------------------------------------------------------------

#--Example 4. Accessing a function AND importing multiple (two) modules

def doit(n):
    from math import sqrt
    from random import random
    return sqrt(random()) + n

t = timeit.Timer('doit(5)',"from __main__ import doit")
print ("13. Ans  =", t.timeit())
#------------------------------------------------------------------------------

#--Example 5. For super accuracy, repeat timings using the repeat function. This increases the default number
#  of iterations, and takes the minimum of the averages times.

t = timeit.Timer( '137.1+13.2')
print ("14. Ans  =", min(t.repeat(3, 2000000))) # default is 1000000
#=============================== END OF PROGRAM ===============================
